/*
  LEGO National Mall Lighting Control System (NaMaLiCS)

  Code for controlling Brickstuff lights using the BrickPixel(tm)
     Arduino shield with BrickPixel(tm) adapter boards
  (c)2021, Bricktectonics

  Portions of this code are based on code developed by
    Brickstuff, Enthusiast Enterprises, LLC. [https://www.brickstuff.com]
    and Adafruit Industries, LLC. [https://www.adafruit.com]

  THIS CODE IS PROVIDED AS-IS WITH NO WARRANTY OR GUARANTEE OF ANY KIND

  HISTORY
  DD.MM.YYYY
  24.10.2021  Version 1.0   Created by Wayne R. Tyler
  29.10.2021  Version 1.0A  Created by Wayne R. Tyler
              Modified for NaMaLiCS tutorial on
              Brickstuff.com forums

*/

// = = = = = = = = = = = = = = = = = = = = = = = = =
// NOTE: A BrickPixel pixel can refer to either an
//       individual Brickstuff light or string of
//       Brickstuff lights.
// = = = = = = = = = = = = = = = = = = = = = = = = =

// -------- LIBRARIES --------------
#include <Adafruit_NeoPixel.h>

// -------- CONSTANTS (won't change) --------------

// The BrickPixel Arduino shield send control commands
//  out on pins 7 & 8.  On the shield pin 7 is labeled
//  PIXEL 1 and pin 8 is labeled PIXEL 2.  The NaMaLiCS
//  testbed uses PIXEL 1.
#define PIN       7

// Minimum and maximum brightness values of a BrickPixel pixel
#define minBrightness 0
#define preferredBrightness 50
#define maxBrightness 255

/*
  The AdaFruit software is designed to work with a NeoPixel light which is
    actually a triplet of LEDs (R,G & B).  The BrickPixel hardware is
    designed to work with individual single color pixels. One Adafruit
    light is the same as three BrickPixel pixels.

  The translation between the two systems is handled by the NUMPIXELS
    definition.

  To calculate the NUMPIXELS, take the total number of BrickPixel pixels
    and divide by 3. If there is any remainder, increase NUMPIXELS by 1.

  For example, if there are 9 Brickstuff pixels attached, the calculation is
    9/3 = 3 with no remainder. So, 3 for the quotient + 0 for the remainder.
    Which means the value of NUMPIXELS is 3.

  REMEMBER: the number of BrickPixel pixels may not be the same as the
    number of Brickstuff lights, since it is possible to have a bank of
    Brickstuff lights connected to single BrickPixel pixel.
*/
#define NUMPIXELS 3

// Declare a NeoPixel object. This is used to control the BrickPixel pixels.
Adafruit_NeoPixel pixels ( NUMPIXELS, PIN, NEO_RGB + NEO_KHZ800 );

// Argument 1 = Number of triplet groups of pixels
// Argument 2 = Arduino pin number of the BrickPixel socket being used
// Argument 3 = Pixel type flags, add together as needed (from the
//              Adafruit NeoPixel documentation):
//    NEO_GRB     Pixels are wired for GRB bitstream (most NeoPixel products)
//    NEO_RGB     Pixels are wired for RGB bitstream (v1 FLORA pixels, not v2)
//    NEO_RGBW    Pixels are wired for RGBW bitstream (NeoPixel RGBW products)
//    NEO_KHZ800  800 KHz bitstream (most NeoPixel products w/ WS2812 LEDs)
//    NEO_KHZ800  400 KHz (classic 'v1' (not v2) FLORA pixels, WS2811 drivers)

// Stores the value of millis() when the program is started
unsigned long startMillis = 0;

// Interval at which to blink (in milliseconds)
//  This set of blinkIntervals are Random Prime Numbers so the
//  blinks aren't synchronous
const unsigned long blinkInterval[] =
  {710, 371, 563, 281, 809, 151, 640, 173, 440, 599};

// -------- VARIABLES (will change) --------------

// Generally, you should use "unsigned long" for variables that hold time
// The value will quickly become too large for an int to store

// Stores the value of millis() in each iteration of loop()
unsigned long currentMillis = 0;

// Will store last time a pixel was updated
//  Used in blinks and flickers on a BrickPixel 9-LED adapter
unsigned long previousPixelMillis[] = {0, 0, 0, 0, 0, 0, 0, 0, 0};

// -------- SETUP --------------
// Setup is run once when the program is started
void setup()
{
  // Setup the serial monitor for displaying text messages
  Serial.begin(9600);

  // Print a text message to the serial monitor
  //  indicating the program has started
  Serial.println("Starting NaMaLiCS");

  // Initialize the NeoPixel object (this is REQUIRED)
  pixels.begin();

  // Clear out any prior BrickPixel settings
  pixels.clear();

  // Set all pixels to minimum brightness
  for( int i=0; i<NUMPIXELS; i++ )
  {
    uint32_t color = pixels.Color(minBrightness, minBrightness, minBrightness);
    pixels.setPixelColor( i, color );
    pixels.show();
  }

  // Get start timing value used by blinking and flickering pixels
  startMillis = millis();
}

// -------- LOOP --------------

void loop() {
  // Refresh the current time each time through the loop
  //  This is used to calculate the timing needed to trigger
  //  events or the duration of an effect
  currentMillis = millis();
  // keeps track of which pixel is being has the effect
  //  being applied
  uint8_t pixelNumber;
  // fxOffset determines the amount of time between pixel events
  //  within an effect and the amount of time between effects
  uint32_t fxOffset = 100;
  // This keeps track of when a pixel effect will start
  uint32_t fxEffectStart = 0;

  // Have all nine pixels in a BrickPixel 9-LED Controller
  //  blink and the blink effect starts immediately.
  for( pixelNumber = 1; pixelNumber <= 9; pixelNumber++ ) {
    blinkPixel( pixelNumber, fxEffectStart );
  }

  // After 3 seconds of blinking pixels, go to the next pixelEffect
  fxEffectStart += 3000;

  // Turn off all the pixels simultaneously
  for( pixelNumber = 1; pixelNumber <= 9; pixelNumber++ ) {
    toggleOff( pixelNumber, fxEffectStart );
  }

  // After previous pixelEffect is finished, update the
  //  start time for the next effect.  Because all the
  //  pixels had the effect applied simultaneously, the
  //  next effect's start time is just a single fxOffset
  fxEffectStart += 1 * fxOffset;

  // Sequentially turn on each pixel starting at pixel 1
  //  and going to pixel 9
  for( pixelNumber = 1; pixelNumber <= 9; pixelNumber++ ) {
    toggleOn( pixelNumber, fxEffectStart + (pixelNumber * fxOffset) );
  }

  // After previous pixelEffect is finished, update the
  //  start time for the next effect.  Because the effect
  //  was applied to the individual pixels sequentially,
  //  the new start time is number of pixels times the fxOffset.
  //
  // REMEMBER: in the previous for loop, the pixelNumber has
  //  been incremented one more time, which causes the for
  //  condition to fail.  Because of that, the pixelNumber
  //  needs to be decremented by one.
  fxEffectStart += (pixelNumber - 1) * fxOffset;

  // Sequentially turn off each pixel starting at pixel 1
  //  and going to pixel 9
  for( pixelNumber = 1; pixelNumber <= 9; pixelNumber++ ) {
    toggleOff( pixelNumber, fxEffectStart + (pixelNumber * fxOffset) );
  }

  // After previous pixelEffect is finished, update the
  //  start time for the next effect
  fxEffectStart += (pixelNumber - 1) * fxOffset;

  // Sequentially turn on each pixel starting at pixel 1
  //  and going to pixel 9
  for( pixelNumber = 1; pixelNumber <= 9; pixelNumber++ ) {
    toggleOn( pixelNumber, fxEffectStart + (pixelNumber * fxOffset) );
  }

  // After previous pixelEffect is finished, update the
  //  start time for the next effect
  fxEffectStart += (pixelNumber - 1) * fxOffset;

  //  starting at pixel 9 and going to pixel 9
  for( pixelNumber = 1; pixelNumber <= 9; pixelNumber++ ) {
    toggleOff( (10 - pixelNumber), fxEffectStart + (pixelNumber * fxOffset) );
  }

  // Update the state of all the pixels.
  pixels.show();
}

// -------- BLINK EFFECT --------------

// Make a pixel blink

// Values passed into the function are:
//  1) the number of the socket in the pixel chain to which
//     a Brickstuff light or light string is attached
//  2) the amount of delay before the blink effect starts

void blinkPixel( uint8_t plugIdentifier, unsigned long delayEffectStart ) {

  // REMEMBER: BrickPixel starts counting with the number one.  Adafruit
  //  begins with the number zero.  The value of the plugIdentifier
  //  must be reduced by one before calculating the the pixelSet and
  //  pixelNumber.

  // Calculate the equivalent NeoPixel triplet the plug is in
  uint8_t pixelSetNumber = (plugIdentifier - 1) / 3;

  // Calculate the equivalent LED in the NeoPixel triplet the plug
  //  corresponds to
  uint8_t pixelNumber = (plugIdentifier - 1) % 3;

  // Get the current brightness of the pixels in the triplet
  uint32_t color = pixels.getPixelColor( pixelSetNumber );

  // Get the individual brightness values of the three pixels in the triplet
  uint8_t pixel01 = (uint8_t)( (color >> 16) & 0xff ),
          pixel02 = (uint8_t)( (color >> 8) & 0xff ),
          pixel03 = (uint8_t)( color & 0xff) ;

  // Check to see if enough time has elapsed to start the blink effect
  if( currentMillis - startMillis >= delayEffectStart ) {

    // Check to see if enough time as elapsed to change the state of the
    //  pixel from on to off or off to on
    if( currentMillis - previousPixelMillis[plugIdentifier - 1] >=
      blinkInterval[(plugIdentifier - 1) % 9] )
    {

      // Change the state of the three pixels
      if( pixelNumber == 0 ) pixel01 ^= preferredBrightness;
      if( pixelNumber == 1 ) pixel02 ^= preferredBrightness;
      if( pixelNumber == 2 ) pixel03 ^= preferredBrightness;

      // Update the pixel triplet with the new values
      color = pixels.Color( pixel01, pixel02, pixel03 );
      pixels.setPixelColor( pixelSetNumber, color );

      // Reset the blink counter
      previousPixelMillis[plugIdentifier - 1] +=
        (blinkInterval[(plugIdentifier - 1) % 9] - delayEffectStart);
    } 
  }
}

// -------- TOGGLE EFFECT --------------

// Toggle a pixel to preferred brightness

// Values passed into the function are:
//  1) the number of the socket in the pixel chain to which
//     a Brickstuff light or light string is attached
//  2) the amount of delay before the toggle effect starts

void toggleOn( uint8_t plugIdentifier, unsigned long delayEffectStart ) {

  // Calculate the equivalent NeoPixel triplet the plug is in
  uint8_t pixelSetNumber = (plugIdentifier - 1) / 3;

  // Calculate the equivalent LED in the NeoPixel triplet the plug
  //  corresponds to
  uint8_t pixelNumber = (plugIdentifier - 1) % 3;

  // Get the current brightness of the pixels in the triplet
  uint32_t color = pixels.getPixelColor( pixelSetNumber );

  // Get the individual brightness values of the three pixels in the triplet
  uint8_t pixel01 = (uint8_t)((color >> 16) & 0xff),
          pixel02 = (uint8_t)((color >> 8) & 0xff),
          pixel03 = (uint8_t)(color & 0xff);

  // Check to see if enough time has elapsed to start the toggle effect
  if( currentMillis - startMillis >= delayEffectStart ) {

    // Check each pixel to see whether it's at minimum brightness
    //  if so, set it to maximum brightness
    if( (pixelNumber == 0) && (pixel01 < preferredBrightness) )
      pixel01 = preferredBrightness;
    if( (pixelNumber == 1) && (pixel02 < preferredBrightness) )
      pixel02 = preferredBrightness;
    if( (pixelNumber == 2) && (pixel03 < preferredBrightness) )
      pixel03 = preferredBrightness;
  }

    // Update the pixel triplet with the new values
    color = pixels.Color( pixel01, pixel02, pixel03 );
    pixels.setPixelColor( pixelSetNumber, color );
}

// = = = = = = = = = = = = = = = = = = = =

// Toggle a pixel to minimum brightness

// Values passed into the function are:
//  1) the number of the socket in the pixel chain to which
//     a Brickstuff light or light string is attached
//  2) the amount of delay before the toggle effect starts

void toggleOff( uint8_t plugIdentifier, unsigned long delayEffectStart ) {

  // Calculate the equivalent NeoPixel triplet the plug is in
  uint8_t pixelSetNumber = (plugIdentifier - 1) / 3;

  // Calculate the equivalent LED in the NeoPixel triplet the plug
  //  corresponds to
  uint8_t pixelNumber = (plugIdentifier - 1) % 3;

  // Get the current brightness of the pixels in the triplet
  uint32_t color = pixels.getPixelColor( pixelSetNumber );

  // Get the individual brightness values of the three pixels in the triplet
  uint8_t pixel01 = (uint8_t)((color >> 16) & 0xff),
          pixel02 = (uint8_t)((color >> 8) & 0xff),
          pixel03 = (uint8_t)(color & 0xff);

  // Check to see if enough time has elapsed to start the toggle effect
  if( currentMillis - startMillis >= delayEffectStart ) {

    // Check each pixel to see whether it's brighter than the minimum
    //  brightness, if so, set it to minimum brightness
    if( (pixelNumber == 0) && (pixel01 > minBrightness) )
      pixel01 = minBrightness;
    if( (pixelNumber == 1) && (pixel02 > minBrightness) )
      pixel02 = minBrightness;
    if( (pixelNumber == 2) && (pixel03 > minBrightness) )
      pixel03 = minBrightness;
  }

  // Update the pixel triplet with the new values
  color = pixels.Color( pixel01, pixel02, pixel03 );
  pixels.setPixelColor( pixelSetNumber, color );
}
